program main

! =====================================================
!
!    MONET (Matrix Ordering algorithm for minimum NET-cuts)
!
!      Yifan Hu, Daresbury Laboratory
! 
!       version: 1.00, Nov. 1998
!
!     Bug report to: y.f.hu@dl.ac.uk
!
! ======================================================

!
!  INPUT: a unsymmetric matrix
! OUTPUT: the column and row ordering to order the matrix into
!         bordered block diagonal form.
!
!
!  Usage
!
!     either : monet <matrix_file> <num_block> <number_of_level (optional)>"
!              where <matrix_file> can be in 
!                   Harwell-Boeing format, in the form *.rua,
!              or 
!                   the MatrixMarket coordinate format, in the form *.mtx
!
!         or : as a subroutine call
!
!                   call monet(n,nz,irn,jcn,num_blocks,row_order, &
!                              cntrl,column_order,column_sta)
!
!        where the last 2 arguments are optional
!               
!



  use readarg_mod
  use basics
  use coo2hb_mod
  use monet_mod

  ! the file to be read and the file to write the row ordering to 
  integer (kind=myint) :: input_unit
  integer (kind = myint) :: output_unit

  ! the input format: 1 coo (i,j,val), 2 harwell-boeing
  integer (kind = myint) :: input_format

  ! number of blocks
  integer (kind = myint) :: num_blocks


  ! control parameter (optional)
  ! 1) moderate (cntrl(1) <= 0) or aggressive (cntrl(1) > 0) -- default 0.0
  ! 2) load imbalance allowed at each bisection (a value in [0,1),
  !    0 for strict balance, 1 for no balance constrain) -- default 0.0
  ! 3) print option: -1 for quiet, 0 for verbose
  ! 4) number of multlevels, 0 for default (10) and positive integers
  !    for user specified levels
  ! 5) not used
  real (kind = myreal), dimension (:) :: cntrl(5)

  ! row_sta: row  starting index for each block
  integer (kind = myint), pointer, dimension (:) :: row_sta

  ! column order: column_order(i) is the original column index
  ! of the i-column of the reordered matrix
  ! column_position: COLUMN_POSITION(I) the new position in the 
  ! reordered matrix of the i-th original column
  integer (kind = myint), pointer, dimension (:) :: column_order
  integer (kind = myint), pointer, dimension (:) :: column_position

  ! column_sta: the starting column index for each block.
  ! column_sta(i) is the starting column index of the i-th block.
  integer (kind = myint), pointer, dimension (:) :: column_sta

  ! binary input?
  logical :: binary

  ! the matrix and right-hand-side.
  integer (kind=myint) :: nz
  integer (kind=myint) :: n,m
  integer (kind=myint), pointer :: irn(:)
  integer (kind=myint), pointer :: jcn(:)
  real (kind = myreal), pointer :: val(:)
  real (kind = myreal), pointer :: rhs(:)

  ! row ordering: row_order(i) is the original row index
  ! of the i-row of the reordered matrix
  ! row_position: ROW_POSITION(I) the new position in the 
  ! reordered matrix of the i-th original row
  integer (kind = myint), pointer, dimension (:) :: row_position
  integer (kind = myint), pointer :: row_order(:)


  integer (kind = myint) :: i,j

  ! maximum size of a block
  integer (kind = myint) :: max_bk_size

  ! firstc: first character of a line
  character (len = 1) :: firstc

  binary = .false.

  ! read in the arguments: matrix file name, num_blocks
  call readarg(input_unit, output_unit, binary, input_format, num_blocks)

  if (input_format == 1) then
     ! read in the matrix and right hand side
     if (binary) then
        read(input_unit) n,nz
        allocate(irn(nz),jcn(nz),val(nz),rhs(n))
        read(input_unit) irn
        read(input_unit) jcn
        read(input_unit) val
     else
        ! get rid of comments in the form %...
        firstc = "%"
        do while (firstc(1:1) == "%")
           read(input_unit,"(a)") firstc
        end do
        backspace(input_unit)
        read(input_unit,*) n,m,nz
        allocate(irn(nz),jcn(nz),val(nz),rhs(n))
        do i = 1, nz
!           read(input_unit,*) irn(i),jcn(i),val(i)
           read(input_unit,*) irn(i),jcn(i)
        end do
        val = 1
        rhs = 1
        m = maxval(jcn)
!        do i = 1, n
!           read(input_unit,*) rhs(i)
!        end do
     end if
     close(input_unit)
  else if (input_format == 2) then
     ! read directly in HB format
     call readhb(input_unit,n,nz,irn,jcn,val,rhs,m)
  else
     stop "this format is not supported"
  end if

  ! allocate space for row and ordering
  allocate(row_order(n))

  ! strict balance of row size in bisection + morderate refinement
  cntrl(1) = 0.0_myreal
  cntrl(2) = 0.0_myreal

  ! 20% unbalance allowed in each bisection, and aggressive refinement!
!  cntrl(1) = 1.0_myreal
!  cntrl(2) = 0.20_myreal
  cntrl(2) = 0.01_myreal


  ! quiet working
  cntrl(3) = -1.0_myreal


  ! default number of levels (huge(0))
  cntrl(4) = 0.0_myreal

  ! the row  starting index for each block 
  allocate(row_sta(num_blocks+1))

  write(*,'(a)') "calling the ordering routine ----------"

  allocate(column_order(m), column_sta(num_blocks+1))
  call monet(n,nz,irn,jcn,num_blocks,row_order, &
       cntrl,row_sta,column_order,column_sta)
  call monet_summarise()

  write(*,'(a)') "finish the ordering routine ----------"

  ! dump block sizes
  do i = 1, num_blocks
     write(*,"('block ',i4,' = ',i10,' X ')") i,row_sta(i+1)-row_sta(i)
  end do

  ! workout the imbalance
  max_bk_size = -huge(0)
  do i = 1, num_blocks
     max_bk_size = max(max_bk_size,row_sta(i+1)-row_sta(i)) 
  end do
  write(*,"(a,f10.2,'%')") "imbalance = ", &
       100*(max_bk_size-real(n)/num_blocks)/(real(n)/num_blocks)

  
  ! if need to reorder:
  allocate(column_position(m), row_position(n))
  do i = 1, m
     column_position(column_order(i)) = i
  end do
  do i = 1, n
     row_position(row_order(i)) = i
  end do
  call dump_coo_gnuplot(n,nz,irn,jcn,val,"original_matrix")
  irn = row_position(irn)
  jcn = column_position(jcn)
  call dump_coo_gnuplot(n,nz,irn,jcn,val,"ordered_matrix")
  write(output_unit,"(a)") "% nrow ncol"
  write(output_unit,"(a)") "% new_row old_row row_block_index"
  write(output_unit,"(a)") "% ...."
  write(output_unit,"(a)") "% new_col old_col col_block_index"
  write(output_unit,"(a)") "% ...."
  do i = 1, num_blocks
     do j = row_sta(i), row_sta(i+1)-1
        write(output_unit,"(i10,i10,i10)") j,row_order(j),i
     end do
  end do
  write(*,*) "column_sta = ",column_sta
  do i = 1, num_blocks
     do j = column_sta(i), column_sta(i+1)-1
        write(output_unit,"(i10,i10,i10)") j,column_order(j),i
     end do
  end do
  do j = column_sta(num_blocks+1),m
     write(output_unit,"(i10,i10,i10)") j,column_order(j),num_blocks+1
  end do
  close(output_unit)

end program main

